import crayons


class Color():

    @staticmethod
    def for_status(status):
        colors_map = {
            # ERRORS
            # 1 == red
            'CREATE_FAILED': crayons.red,
            'DELETE_FAILED': crayons.red,
            'UPDATE_FAILED': crayons.red,
            'ROLLBACK_IN_PROGRESS': crayons.red,
            'ROLLBACK_FAILED': crayons.red,
            'UPDATE_ROLLBACK_FAILED': crayons.red,

            # COMPLETE
            # 2 == green
            'ROLLBACK_COMPLETE': crayons.green,
            'CREATE_COMPLETE': crayons.green,
            'DELETE_COMPLETE': crayons.green,
            'UPDATE_COMPLETE': crayons.green,
            'UPDATE_ROLLBACK_COMPLETE': crayons.green,

            # SUCCESS
            # 3 == yellow
            'CREATE_IN_PROGRESS': crayons.yellow,
            'DELETE_IN_PROGRESS': crayons.yellow,
            'UPDATE_IN_PROGRESS': crayons.yellow,
            'UPDATE_ROLLBACK_IN_PROGRESS': crayons.yellow,
            'UPDATE_COMPLETE_CLEANUP_IN_PROGRESS': crayons.yellow,
            'UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS': crayons.yellow,

            # SKIPPED
            # 8 == grey
            'DELETE_SKIPPED': crayons.cyan,
        }
        return '{}'.format(colors_map[status](status))



"""
Configuration.

"""

import os

import click
import delegator
import yaml
import jinja2


def is_installed(cmd):
    """Check that ``cmd`` is installed and available in $PATH."""
    c = delegator.run([cmd])
    if c.err:
        return False
    return True


def is_git_repo():
    """Check that the current directory is a functioning git repository."""
    c = delegator.run('git status')
    if 'fatal: Not a git repository' in c.err:
        return False
    return True


class Config():

    @staticmethod
    def env(key, default=None):
        """Return the value of the `key` environment variable."""
        try:
            if default:
                return os.getenv(key, default)
            return os.environ[key]
        except KeyError:
            click.secho('[ERROR] No environment variable with key {}'.format(key), err=True, fg='red')
            exit(1)

    @staticmethod
    def _git_commit_msg():
        """
        Return the message (subject) of the latest Git commit.

        YAML complains if the commit message contains single quotes, so we
        remove those.
        """
        c = delegator.run('git log -1 --pretty=%s')
        return c.out.strip().replace('\'', '')

    @staticmethod
    def _git_commit():
        """Return the SHA1 of the latest Git commit (HEAD)."""
        c = delegator.run('git rev-parse --short HEAD')
        return c.out.strip()

    @staticmethod
    def _git_branch():
        """Return the name of the current Git branch."""
        c = delegator.run('git rev-parse --abbrev-ref HEAD')
        return c.out.strip()

    @staticmethod
    def git_config():
        if not is_installed('git'):
            click.secho('[WARN] git is not installed or not in $PATH', err=True, fg='red')
            return {}
        if not is_git_repo():
            click.secho('[ERROR] Current directory is not a Git repository', err=True, fg='red')
            return {}
        return dict(
            branch_name=Config._git_branch(),
            commit_sha1=Config._git_commit(),
            commit_msg=Config._git_commit_msg()
        )

    @staticmethod
    def load(config_file):
        """
        Return the YAML configuration for a project based on the `config_file` template.

        By default, the template exposes the `env` function.
        The `git_branch` and `git_commit` values are exposed only when a `.git` folder
        exists in the current directory
        """
        template = Config.render(config_file)
        template_env = dict(
            env=Config.env,
            git=Config.git_config(),
            git_branch=Config._git_branch(),
            git_commit=Config._git_commit())
        try:
            return yaml.load(template.render(**template_env))
        except jinja2.exceptions.UndefinedError as err:
            click.secho('[ERROR] {0} in {1}'.format(err.message, config_file), err=True, fg='red')
            exit(1)

    @staticmethod
    def render(config_file):
        path, filename = os.path.split(os.path.abspath(config_file))
        try:
            return jinja2.Environment(
                loader=jinja2.FileSystemLoader(path or './'),
                undefined=jinja2.StrictUndefined).get_template(filename)
        except jinja2.exceptions.TemplateNotFound:
            click.secho('[ERROR] No such file or directory: {0}'.format(config_file), err=True, fg='red')
            exit(1)



"""
Stack.
"""

import time
from datetime import datetime, timedelta

import click
import crayons
import pytz
from botocore.exceptions import ClientError

from brume.boto_client import cfn_client
from brume.color import Color

TZ = pytz.timezone('UTC')


def _print_log_headers():
    click.echo('{:23s} {:36s} {:30s} {:30s} {}'.format(
        'Timestamp', 'Status', 'Resource', 'Type', 'Reason'
    ))


def _log_event(event):
    click.echo('{:23s} {:36s} {:30s} {:30s} {}'.format(
        event['Timestamp'].strftime('%Y-%m-%d %H:%M:%S UTC'),
        Color.for_status(event['ResourceStatus']),
        event['LogicalResourceId'],
        event['ResourceType'],
        event.get('ResourceStatusReason', ''),
    ))


def _make_tags(tags_list):
    return [{'Key': k, 'Value': v} for k, v in tags_list.items()]


def _make_parameters(params_list):
    return [{'ParameterKey': k, 'ParameterValue': v} for k, v in params_list.items()]


def _outputs_for(stack):
    try:
        outputs = cfn_client().describe_stacks(StackName=stack)['Stacks'][0].get('Outputs', {})
        return {o['OutputKey']: o['OutputValue'] for o in outputs}
    except ClientError as e:
        if 'does not exist' in e.message:
            click.secho('Stack [{0}] does not exist'.format(stack), err=True, fg='red')
            exit(1)
        else:
            raise e


class Stack(object):
    """
    Represent a CloudFormation stack.
    """

    stack_name = None
    capabilities = []

    def __init__(self, conf):
        self.stack_name = conf['stack_name']
        self.template_body = conf['template_body']
        self.on_failure = conf.get('on_failure', 'ROLLBACK')
        self.capabilities = conf.get('capabilities', [])
        self.parameters = _make_parameters(conf.get('parameters', {}))
        self.tags = _make_tags(conf.get('tags', {}))

        # Check the events 30 seconds before if the stack update starts way too soon
        self.update_started_at = datetime.now(TZ) - timedelta(seconds=30)

    @property
    def configuration(self):
        return dict(
            StackName=self.stack_name,
            TemplateBody=open(self.template_body, 'r').read(),
            Parameters=self.parameters,
            Capabilities=self.capabilities,
            Tags=self.tags,
        )

    def get_stacks(self):
        """
        Return a list of stacks containing the current stack and its nested stack resources.
        """
        stacks = [self.stack_name]
        substacks = cfn_client().describe_stack_resources(StackName=self.stack_name)
        stacks.extend([
            s['PhysicalResourceId']
            for s in substacks['StackResources']
            if s['ResourceType'] == 'AWS::CloudFormation::Stack'
        ])
        return stacks

    def outputs(self):
        """
        Return a dict containing the outputs of the current stack and its nested stacks.
        """
        outputs = {self.stack_name: _outputs_for(self.stack_name)}
        stack_resources = cfn_client().describe_stack_resources(StackName=self.stack_name)
        outputs.update({
            res['LogicalResourceId']: _outputs_for(res['PhysicalResourceId'])
            for res in stack_resources['StackResources']
            if res['ResourceType'] == 'AWS::CloudFormation::Stack'
        })
        return outputs

    def params(self):
        """
        Return a dict containing the parameters of the current stack and its nested stacks.
        """
        try:
            return {
                stack: {
                    param['ParameterKey']: param['ParameterValue']
                    for param in cfn_client().describe_stacks(StackName=stack)['Stacks'][0].get('Parameters', [])
                } for stack in self.get_stacks()
            }
        except ClientError as e:
            if 'does not exist' in e.message:
                click.secho('Stack [{0}] does not exist'.format(self.stack_name), err=True, fg='red')
                exit(1)
            else:
                raise e

    @staticmethod
    def exists(stack_name):
        """
        Return `True` if a stack exists with the name `stack_name`, `False` otherwise.
        """
        try:
            cfn_client().describe_stacks(StackName=stack_name)
        except ClientError as e:
            if 'does not exist' in e.message:
                return False
        else:
            return True

    def create(self):
        """
        Create the stack in CloudFormation.
        """
        click.echo('Creating stack {0}...'.format(self.stack_name))
        try:
            cfn_client().create_stack(**self.configuration)
            time.sleep(5)
            self.tail()
        except ClientError as err:
            if 'AlreadyExistsException' in err.message:
                click.secho('Stack [{0}] already exists'.format(self.stack_name), err=True, fg='red')
                exit(1)
            else:
                click.secho(err.message, err=True, fg='red')

    def update(self):
        """
        Update the stack in CloudFormation if it exists.
        """
        click.echo('Updating stack {0}...'.format(self.stack_name))
        try:
            cfn_client().update_stack(**self.configuration)
            self.tail()
        except ClientError as err:
            if 'does not exist' in err.message:
                click.secho('Stack [{0}] does not exist'.format(self.stack_name), err=True, fg='red')
                exit(1)
            elif 'No updates are to be performed.' in err.message:
                click.echo(crayons.yellow('No updates are to be performed on stack [{}]'.format(
                    self.stack_name)), err=True)
                exit(1)
            else:
                click.secho(err.message, err=True, fg='red')

    def create_or_update(self):
        """
        Create or update the stack in CloudFormation if it already exists.
        """
        if Stack.exists(self.stack_name):
            self.update()
        else:
            self.create()

    def delete(self):
        """
        Delete the stack in CloudFormation.
        """
        click.echo('Deleting stack {0}...'.format(self.stack_name))
        if Stack.exists(self.stack_name):
            cfn_client().delete_stack(StackName=self.stack_name)
            try:
                self.tail(catch_error=True)
                click.echo(crayons.yellow('Stack [{0}] is deleted'.format(self.stack_name)))
            except ClientError as err:
                raise err
        else:
            click.secho('Stack [{0}] does not exist'.format(self.stack_name), err=True, fg='red')
            exit(1)

    def status(self):
        """
        Return the status of the stack in CloudFormation, based on the last stack event.
        """
        try:
            stacks = cfn_client().describe_stacks(StackName=self.stack_name)
            click.echo(Color.for_status(next(s['StackStatus'] for s in stacks['Stacks'])))
        except KeyError as err:
            click.secho(err, err=True, fg='red')
            exit(1)
        except ClientError as err:
            if 'does not exist' in err.message:
                click.secho('Stack [{}] does not exist'.format(self.stack_name), err=True, fg='red')
                exit(1)

    def get_events(self):
        """
        Fetch the stack events
        """
        events = cfn_client().describe_stack_events(StackName=self.stack_name)
        return reversed(events['StackEvents'])

    def tail(self, sleep_time=3, catch_error=False):
        """
        Tail the event log of the stack.
        """
        error = False
        seen = set()
        try:
            events = self.get_events()
            _print_log_headers()
            while True:
                for event in events:
                    if event['Timestamp'] < self.update_started_at:
                        seen.add(event['EventId'])
                    if event['EventId'] in seen:
                        continue
                    if 'FAILED' in event['ResourceStatus']:
                        error = True
                    _log_event(event)
                    seen.add(event['EventId'])
                if self.stack_complete(event):
                    if error:
                        exit(1)
                    break
                time.sleep(sleep_time)
                events = self.get_events()
        except ClientError as err:
            if 'does not exist' in err.message and catch_error:
                return False
            else:
                raise err

    def stack_complete(self, event):
        """
        Return True if the stack has reached a COMPLETE status.
        """
        return (
            event['LogicalResourceId'] == self.stack_name and
            event['ResourceStatus'].endswith('COMPLETE')
        )



"""
Template module.
"""

import logging
import os

import click
import crayons
from botocore.exceptions import ClientError

from brume.boto_client import cfn_client, s3_client

logging.getLogger('botocore').setLevel(logging.WARNING)

CFN_TEMPLATE_SIZE_LIMIT = 51200


class Template(object):
    """
    CloudFormation template.
    """

    key = None

    def __init__(self, file_path, config):
        self.local_file_path = file_path
        self.file_path = file_path
        self.s3_bucket = config['s3_bucket']
        self.s3_path = config.get('s3_path', '')
        local_path = config.get('local_path', '')
        if local_path != '.':
            self.file_path = self.file_path.replace(local_path, '')

    @property
    def public_url(self):
        s3_url = os.path.normpath('{0}.s3.amazonaws.com/{1}'.format(self.s3_bucket, self.s3_key))
        return 'https://{0}'.format(s3_url)

    @property
    def s3_key(self):
        return os.path.normpath('{0}/{1}'.format(self.s3_path, self.file_path)).strip('/')

    @property
    def size(self):
        return os.path.getsize(self.local_file_path)

    @property
    def template_is_too_large(self):
        return self.size > CFN_TEMPLATE_SIZE_LIMIT

    @property
    def content(self):
        try:
            with open(self.local_file_path, 'r') as _file:
                return _file.read()
        except IOError as err:
            click.echo(crayons.red('File {!r} does not exist').format(self.local_file_path), err=True)
            raise err

    def validate(self):
        """
        Validate the template on CloudFormation.

        If the template is larger than CFN_TEMPLATE_SIZE_LIMIT, brume uploads a
        copy of the template with the .copy suffix on S3 and performs validation
        on this template.
        """
        validation_path = self.local_file_path
        params = {'TemplateBody': self.content}
        if self.template_is_too_large:
            # Template will be copied, uploaded and validated on S3
            self.upload(copy=True)
            validation_path = self.public_url + '.copy'
            params = {'TemplateURL': validation_path}
        try:
            click.echo('Validating {0} ...'.format(crayons.yellow(validation_path)), nl=False)
            cfn_client().validate_template(**params)
        except ClientError as error:
            click.echo(crayons.red('invalid'))
            click.echo(error.message, err=True)
            return False
        click.echo(crayons.green('valid'))
        return True

    def upload(self, copy=False):
        """
        Upload the template to S3.

        If copy is True, uploads a copy of the template with the .copy suffix.
        """
        s3_key = self.s3_key
        public_url = self.public_url
        if copy:
            s3_key += '.copy'
            public_url += '.copy'
        click.echo('Publishing {0} to {1}'.format(crayons.yellow(self.local_file_path), public_url))
        s3_client().put_object(Bucket=self.s3_bucket, Body=self.content, Key=s3_key)
        return self



"""
cfn-check
"""
import json
import logging
import os
import sys

import click
import crayons

LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

CFN_REF = 'Ref'
CFN_GETATT = 'Fn::GetAtt'


class Stack(object):
    """docstring for Stack."""

    def __init__(self, name):
        self.name = name
        self.outputs = {}
        self.resources = {}
        self.template_url = ''

        # Parameters that are passed when declaring the stack as a resource
        # of the main stack
        self.input_parameters = {}

        # Parameters defined in the Parameters section of the template
        self.parameters = {}

    def find(self, key):
        """
        Return a list of resources and outputs that contain ``key``.
        """
        outputs_refs = list(Stack.find_nodes(self.outputs, key))
        resources_refs = list(Stack.find_nodes(self.resources, key))
        return resources_refs + outputs_refs

    def missing_refs(self):
        """
        Return a list of Ref statements that point to non-existing resources or
        parameters in the template.
        """
        return [ref for ref in self.find(CFN_REF) if not self.has_ref(ref)]

    def has_ref(self, ref):
        """
        Check that ``ref`` points to a resource that is actually declared in the
        template or in the parameters of the stack.
        """
        return ref in self.resources or ref in self.parameters

    def missing_getatt(self):
        """
        Return a list of GetAtt statements that point to non-existing resources
        or parameters in the template.
        """
        return [ref for ref in self.find(CFN_GETATT) if not self.has_getatt(ref)]

    def has_getatt(self, getatt):
        """
        Check that ``getatt`` points to a resource that is actually declared in
        the template or in the parameters of the stack.
        """
        return getatt[0] in self.resources or getatt[0] in self.parameters

    def missing_parameters(self):
        """
        Return a list of parameters that are expected in the substack with no
        default value but have no parameter in the main stack.
        """
        return [
            param_name
            for param_name, param in self.parameters.items()
            if param_name not in self.input_parameters and 'Default' not in param
        ]

    def extra_parameters(self):
        """
        Return a list of parameters that are passed from the main stack but are
        not expected in the substack definition.
        """
        return [
            param_name
            for param_name, _ in self.input_parameters.items()
            if param_name not in self.parameters
        ]

    def load_from_file(self, name=None):
        """
        Create a Stack from its template.
        """
        stack_name = name if name else self.name
        try:
            template = open(stack_name, 'r')
        except IOError as err:
            click.echo('Template for stack {0} not found'.format(stack_name), err=True)
            click.echo(err, err=True)
            sys.exit(1)
        template = json.load(template)
        self.outputs = template.get('Outputs', {})
        self.parameters = template.get('Parameters', {})
        self.resources = template.get('Resources', {})

    @staticmethod
    def new_substack(stack_name, resource):
        """
        Create a substack.
        """
        newstack = Stack(stack_name)
        newstack.input_parameters = resource['Properties'].get('Parameters', {})
        newstack.template_url = resource['Properties'].get('TemplateURL', {})
        return newstack

    @staticmethod
    def aws_pseudo_parameter(v):
        """
        Check that ``v`` is an AWS pseudo-parameter (like AWS::Region).
        """
        return isinstance(v, (str, unicode)) and v.startswith('AWS::')

    @staticmethod
    def find_nodes(node, key):
        if not isinstance(node, dict):
            return
        for k, v in node.items():
            if Stack.aws_pseudo_parameter(v):
                continue
            if k == key:
                yield v
            elif isinstance(v, list):
                for n in v:
                    for id_val in Stack.find_nodes(n, key):
                        yield id_val
            elif isinstance(v, dict):
                for id_val in Stack.find_nodes(v, key):
                    yield id_val


def check_templates(template):
    """
    Checks:
    - that there are parameters sent from the main stack for every expected
      parameter in the substacks
    - that `Ref` and `GetAtt` point to existing resources or parameters in the
      template
    - that the outputs used as params in the main stack exist in the substacks
    """
    templates_path, filename = os.path.split(os.path.realpath(template))
    main_stack_name, filetype = os.path.splitext(filename)

    main_stack = Stack(main_stack_name)
    main_stack.load_from_file(os.path.join(templates_path, filename))

    stacks = {
        name: Stack.new_substack(name, resource)
        for name, resource in main_stack.resources.items()
        if resource['Type'] == 'AWS::CloudFormation::Stack'
    }
    stacks[main_stack_name] = main_stack

    error = False
    for name, substack in stacks.items():
        substack_path = os.path.join(templates_path, name) + filetype
        LOGGER.debug('Loading Stack %s file %s', name, substack_path)
        substack.load_from_file(substack_path)

        if name != main_stack_name:
            # We don't validate Parameters on the Main stack
            for param in substack.missing_parameters():
                click.echo('Stack {0} should give substack {1} parameter: {2}'.format(
                    crayons.yellow(main_stack_name),
                    crayons.yellow(name),
                    crayons.red(param)
                ), err=True)
                error = True
            for param in substack.extra_parameters():
                click.echo('Stack {0} is giving extra parameter {1} to substack: {2}'.format(
                    crayons.yellow(main_stack_name),
                    crayons.yellow(name),
                    crayons.red(param)
                ), err=True)
                error = True

        for ref in substack.missing_refs():
            click.echo('Stack {0} has undefined {1} statement: {2}'.format(
                crayons.yellow(name), crayons.yellow('Ref'), crayons.red(ref)
            ), err=True)
            error = True

        for getatt in substack.missing_getatt():
            click.echo('Stack {0} has undefined {1} statement: {2}'.format(
                crayons.yellow(name), crayons.yellow('GetAtt'), crayons.red(getatt)
            ), err=True)
            error = True

    # Special case for the Main stack GetAtt
    for att in main_stack.find(CFN_GETATT):
        if att[1].startswith('Outputs.') and att[0] in stacks:
            stack = stacks[att[0]]
            output_name = att[1].replace('Outputs.', '')
            if output_name not in stacks[att[0]].outputs:
                click.echo('Stack {0} references undefined Output {1} from substack {2}'.format(
                    crayons.yellow(main_stack_name),
                    crayons.red(output_name),
                    crayons.yellow(stack.name)
                ), err=True)
                error = 1

    if not error:
        click.echo(crayons.green('Congratulations, your templates appear to be OK!\n'))
    else:
        sys.exit(error)



"""
Assets that will be uploaded to S3.

Assets can be userdata, scripts, configuration files and more.
"""

import os

import click
import crayons

from brume.boto_client import s3_client


def send_assets(local_path, s3_bucket, s3_path=''):
    """
    Send directory '{local_path}' under 's3://{s3_bucket}/{s3_path}'.
    """
    for (dirpath, _dirnames, filenames) in os.walk(local_path):
        for filename in filenames:
            sourcepath = os.path.join(dirpath, filename)
            key = s3_path + '/' + os.path.relpath(sourcepath, local_path)
            click.echo('Publishing {} to {}'.format(
                crayons.yellow(sourcepath), s3_bucket + '/' + key))
            with click.open_file(sourcepath) as asset:
                s3_client().put_object(
                    Bucket=s3_bucket, Body=asset.read(), Key=key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Setup module.
"""

import io
import os
import re
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

from brume import VERSION

# Package meta-data.
NAME = 'brume'
DESCRIPTION = 'AWS Cloudformation deployer. '
URL = 'https://github.com/flou/brume'
AUTHORS = {
    'ferrand@ekino.com': 'Lou Ferrand',
}

# What packages are required for this module to be executed?
REQUIRED = [
    'boto3>=1.4.5',
    'crayons==0.1.2',
    'click>=6.7',
    'PyYAML==3.12',
    'Jinja2==2.9.6',
    'pytz==2017.2',
    'delegator.py==0.0.13',
]

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

HERE = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
with io.open(os.path.join(HERE, 'README.rst'), encoding='utf-8') as f:
    LONG_DESCRIPTION = '\n' + f.read()


def _download_url():
    return '{repo}/repository/archive.tar.gz?ref={version}'.format(
        repo=URL, version=VERSION
    )


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


class PublishCommand(Command):
    """Support setup.py publish."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(HERE, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        sys.exit()


setup(
    name='brume',
    version=re.sub(r'[^\d\.]', '', VERSION),
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url=URL,
    author=', '.join(AUTHORS.values()),
    author_email=', '.join(AUTHORS.keys()),
    license='MIT',
    packages=find_packages(exclude=['tests', 'cover']),
    include_package_data=True,
    download_url=_download_url(),
    zip_safe=False,
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'moto'],
    keywords=['AWS', 'CloudFormation'],
    install_requires=[
        'boto3>=1.4.5',
        'crayons==0.1.2',
        'click>=6.7',
        'PyYAML==3.12',
        'Jinja2==2.9.6',
        'pytz==2017.2',
        'delegator.py==0.0.13',
    ],
    entry_points={
        'console_scripts': ['brume=brume.cli:cli']
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    cmdclass={
        'publish': PublishCommand,
    }
)



"""
Brume CLI module.
"""

from glob import glob
from os import path

import click
from yaml import dump
import json

from . import VERSION
from .assets import send_assets
from .checker import check_templates
from .config import Config
from .stack import Stack
from .template import Template

DEFAULT_BRUME_CONFIG = 'brume.yml'

class Context(object):
    """
    Context object to hold the configuration.
    """

    def __init__(self):
        self.config = dict()
        self.stack = None
        self.debug = False


pass_ctx = click.make_pass_decorator(Context, ensure=True)


def config_callback(ctx, _, value):
    ctx = ctx.ensure_object(Context)
    ctx.config = Config.load(value)
    ctx.stack = Stack(ctx.config['stack'])
    return value


@click.group()
@click.version_option(VERSION, '-v', '--version')
@click.help_option('-h', '--help')
@click.option('-c', '--config', expose_value=False, default=DEFAULT_BRUME_CONFIG,
              help='Configuration file (defaults to {}).'.format(DEFAULT_BRUME_CONFIG),
              callback=config_callback)
def cli():
    pass


@cli.command()
@pass_ctx
def config(ctx):
    """
    Print the current stack confguration.
    """
    click.echo(dump(ctx.config, default_flow_style=False))


@cli.command()
@pass_ctx
def create(ctx):
    """
    Create a new CloudFormation stack.
    """
    validate_and_upload(ctx.config)
    ctx.stack.create()


@cli.command()
@pass_ctx
def update(ctx):
    """
    Update an existing CloudFormation stack.
    """
    validate_and_upload(ctx.config)
    ctx.stack.update()


@cli.command()
@pass_ctx
def deploy(ctx):
    """
    Create or update a CloudFormation stack.
    """
    validate_and_upload(ctx.config)
    ctx.stack.create_or_update()
    ctx.stack.outputs()


@cli.command()
@pass_ctx
def delete(ctx):
    """
    Delete the CloudFormation stack.
    """
    ctx.stack.delete()


@cli.command()
@pass_ctx
def status(ctx):
    """
    Get the status of a CloudFormation stack.
    """
    ctx.stack.status()


@cli.command()
@pass_ctx
@click.option('--flat', is_flag=True, help='Flat format (only for the text output format)')
@click.option('output_format', '-f', '--format', type=click.Choice(['text', 'json', 'yaml']),
              default='text', help='Output format (text/json/yaml)')
def outputs(ctx, output_format, flat=False):
    """
    Get the full list of outputs of a CloudFormation stack.
    """
    stack_outputs = ctx.stack.outputs()
    if output_format == 'text':
        if flat:
            for _, outputs in stack_outputs.items():
                for output_name, output_value in outputs.items():
                    click.echo('{} = {}'.format(output_name, output_value))
        else:
            for stack_name, outputs in stack_outputs.items():
                if not outputs:
                    continue
                if ':stack/' in stack_name:
                    stack_name = stack_name.partition(':stack/')[2].split('/')[0]
                click.echo(stack_name)
                for output_name, output_value in outputs.items():
                    click.echo('\t{} = {}'.format(output_name, output_value))
                click.echo()
    elif output_format == 'yaml':
        click.echo(dump(stack_outputs, default_flow_style=False))
    elif output_format == 'json':
        click.echo(json.dumps(stack_outputs, indent=True))


@cli.command()
@pass_ctx
@click.option('--flat', is_flag=True, help='Flat format (only for the text output format)')
@click.option('output_format', '-f', '--format', type=click.Choice(['text', 'json', 'yaml']),
              default='text', help='Output format (text/json/yaml)')
def parameters(ctx, output_format, flat=False):
    """
    Get the full list of parameters of a CloudFormation stack.
    """
    stack_params = ctx.stack.params()
    if output_format == 'text':
        if flat:
            for _, stack_params in stack_params.items():
                for param_name, param_value in stack_params.items():
                    click.echo('{} = {}'.format(param_name, param_value))
        else:
            for stack_name, stack_parameters in stack_params.items():
                if not stack_parameters:
                    continue
                if ':stack/' in stack_name:
                    stack_name = stack_name.partition(':stack/')[2].split('/')[0]
                click.echo(stack_name)
                for param_name, param_value in stack_parameters.items():
                    click.echo('\t{} = {}'.format(param_name, param_value))
                click.echo()
    elif output_format == 'yaml':
        click.echo(dump(stack_params, default_flow_style=False))
    elif output_format == 'json':
        click.echo(json.dumps(stack_params, indent=True))


@cli.command()
@pass_ctx
def validate(ctx):
    """
    Validate CloudFormation templates.
    """
    error = False
    for t in collect_templates(ctx.config):
        valid = t.validate()
        if not valid:
            error = True
    if error:
        exit(1)


@cli.command()
@pass_ctx
def upload(ctx):
    """
    Upload CloudFormation templates and assets to S3.
    """
    process_assets(ctx.config)
    return [t.upload() for t in collect_templates(ctx.config)]


@cli.command()
@pass_ctx
def check(ctx):
    """
    Check CloudFormation templates.
    """
    check_templates(ctx.config['stack']['template_body'])


@cli.command()
def init():
    """
    Initialise a Brume project with a simple configuration file ({}).
    """
    if path.isfile(DEFAULT_BRUME_CONFIG):
        click.echo('{0} already exists in current directory. Nothing to do.'.format(DEFAULT_BRUME_CONFIG))
    else:
        with open(DEFAULT_BRUME_CONFIG, 'w') as brume_file:
            brume_file.write("""
---
region: {{ env('AWS_DEFAULT_REGION', 'eu-west-1') }}

{% set stack_name = 'my-stack' %}

stack:
    stack_name: {{ stack_name }}

    template_body: Main.json
    capabilities: [ CAPABILITY_IAM ]
    on_failure: DELETE

    parameters:
    GitCommit:  '{{ git_commit }}'
    GitBranch:  '{{ git_branch }}'

templates:
    s3_bucket: my_bucket
    s3_path: {{ stack_name }}
""")


def process_assets(conf):
    """
    Upload project assets to S3.
    """
    if 'assets' not in conf:
        return
    assets_config = conf['assets']
    local_path = assets_config['local_path']
    s3_bucket = assets_config['s3_bucket']
    s3_path = assets_config['s3_path']
    click.echo('Processing assets from {} to s3://{}/{}'.format(local_path, s3_bucket, s3_path))
    send_assets(local_path, s3_bucket, s3_path)


def collect_templates(conf):
    """
    Convert every template into a brume.Template.

    The type of the templates is determined based on the `template_body`
    property of the configuration file.
    """
    _file, ext = path.splitext(conf['stack']['template_body'])
    template_paths = glob(path.join(conf['templates'].get('local_path', ''), '*' + ext))
    return [Template(t, conf['templates']) for t in template_paths]


def validate_and_upload(conf):
    """
    Validate and upload CloudFormation templates to S3.
    """
    templates = collect_templates(conf)
    error = False
    for t in templates:
        if not t.validate():
            error = True
    if error:
        exit(1)
    for t in templates:
        t.upload()
    process_assets(conf)


if __name__ == '__main__':
    cli()
